--author: igor29381
--v1.1
--date: 28.07.18

PalletLoader = {};

PalletLoader.STATE_CLOSED = 1;
PalletLoader.STATE_PREPARE = 2;
PalletLoader.STATE_READY = 3;
PalletLoader.STATE_LOAD = 4;
PalletLoader.STATE_BEFOREUNLOAD = 5;
PalletLoader.STATE_UNLOAD = 6;
PalletLoader.STATE_CLOSING = 7;

PalletLoader.SENDSTATE_PREPARE = 1;
PalletLoader.SENDSTATE_CLOSING = 2;
PalletLoader.SENDSTATE_LOAD = 3;
PalletLoader.SENDSTATE_UNLOADING = 4;
PalletLoader.SENDSTATE_GRABMODE = 5;

local hudFile = g_currentModDirectory.."textures/shared/pallets.dds";
FillUtil.registerFillType("pallets", g_i18n:getText("pallets"), nil, 1, false, hudFile, hudFile, 0.85, 0);

function PalletLoader.prerequisitesPresent(specializations)
	return true;
end;

function PalletLoader:load(savegame)
	if hasXMLProperty(self.xmlFile, "vehicle.palletLoader") then
		self.palletLoader = {};
		self.getIsLoaderActive = PalletLoader.getIsLoaderActive;
		self.getPalletFromArea = PalletLoader.getPalletFromArea;
		self.getEmptyPosition = PalletLoader.getEmptyPosition;
		self.getLastUsedPosition = PalletLoader.getLastUsedPosition;
		self.getUnloadingPlaceIsEmpty = PalletLoader.getUnloadingPlaceIsEmpty;
		self.addPalletToPosition = PalletLoader.addPalletToPosition;
		self.removePalletFromPosition = PalletLoader.removePalletFromPosition;
		self.onChangeState = PalletLoader.onChangeState;
		self.getDoConsumePtoPower = PalletLoader.getDoConsumePtoPower;
		self.getPtoRpm = PalletLoader.getPtoRpm;
		local startPointIndex = getXMLString(self.xmlFile, "vehicle.palletLoader#startPointIndex");
		if startPointIndex then
			local startPointNode = Utils.indexToObject(self.components, startPointIndex);
			if startPointNode > 0 then
				self.palletLoader.startPoint = startPointNode;
			end;
		end;
		local endPointIndex = getXMLString(self.xmlFile, "vehicle.palletLoader#endPointIndex");
		if endPointIndex then
			local endPointNode = Utils.indexToObject(self.components, endPointIndex);
			if endPointNode > 0 then
				local x,_,z = getTranslation(endPointNode);
				self.palletLoader.dimensions = {x, z};
			end;
		end;
		local grabPointIndex = getXMLString(self.xmlFile, "vehicle.palletLoader#grabPointIndex");
		if grabPointIndex then
			local grabPointNode = Utils.indexToObject(self.components, grabPointIndex);
			if grabPointNode > 0 then
				self.palletLoader.grabPoint = grabPointNode;
			end;
		end;
		local unloadPointIndex = getXMLString(self.xmlFile, "vehicle.palletLoader#unloadPointIndex");
		if unloadPointIndex then
			local unloadPointNode = Utils.indexToObject(self.components, unloadPointIndex);
			if unloadPointNode > 0 then
				self.palletLoader.unloadPoint = unloadPointNode;
			end;
		end;
		local palletGroupIndex = getXMLString(self.xmlFile, "vehicle.palletLoader#palletGroupIndex");
		if palletGroupIndex then
			local palletGroupNode = Utils.indexToObject(self.components, palletGroupIndex);
			local count = getNumOfChildren(palletGroupNode);
			if palletGroupNode > 0 and count > 0 then
				self.palletLoader.palletGroup = palletGroupNode;
				self.palletLoader.positions = {};
				self.palletLoader.count = count;
				self.palletLoader.usedCount = 0;
				for i=1, count do
					local node = getChildAt(palletGroupNode, i-1);
					local moveablePoint = getChildAt(node, 0);
					local direction = 1;
					local _,remainder = math.modf(i/2);
					if remainder == 0 then
						direction = -1;
					end;
					table.insert(self.palletLoader.positions, {node=node, moveablePoint=moveablePoint, direction=direction});
				end;
				self.palletLoader.startPosition = {getTranslation(palletGroupNode)};
			end;
		end;
		self.palletLoader.lockJointIndex = getXMLInt(self.xmlFile, "vehicle.palletLoader#lockJointIndex");
		self.palletLoader.state = PalletLoader.STATE_CLOSED;
		self.palletLoader.grabMode = false;
		self.palletLoader.dropMode = false;
		self.palletLoader.isFolding = false;
		self.palletLoader.currentMoving = 0;
		self.palletLoader.currentPosition = 0;
		self.palletLoader.dropPosition = self.palletLoader.startPosition[3];
		self.palletLoader.lastDropPosition = self.palletLoader.startPosition[3];
		self.palletLoader.dropDirection = 1;
		self.palletLoader.loadText = {};
		self.palletLoader.loadText[true] = g_i18n:getText("action_stopLoadPallet");
		self.palletLoader.loadText[false] = g_i18n:getText("action_startLoadPallet");
		local foldInputButton = Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.palletLoader#foldInputButton"), "IMPLEMENT_EXTRA2");
		self.palletLoader.foldInputButton = InputBinding[foldInputButton];
		local loadInputButton = Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.palletLoader#loadInputButton"), "IMPLEMENT_EXTRA3");
		self.palletLoader.loadInputButton = InputBinding[loadInputButton];
		local unloadInputButton = Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.palletLoader#unloadInputButton"), "TOGGLE_TIPSTATE");
		self.palletLoader.unloadInputButton = InputBinding[unloadInputButton];
	end;
end;

function PalletLoader:delete()
end;

function PalletLoader:mouseEvent(posX, posY, isDown, isUp, button)
end;

function PalletLoader:keyEvent(unicode, sym, modifier, isDown)
end;

function PalletLoader:postLoad(savegame)
	if savegame ~= nil then
        self.palletLoader.state = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#palletLoaderState"), 1);
		self.palletLoader.loadPallets = {};
        local i=0;
        while true do
            local palletKey = savegame.key..string.format(".pallet(%d)", i);
            if not hasXMLProperty(savegame.xmlFile, palletKey) then
				break;
			end;
            local i3dFilename = getXMLString(savegame.xmlFile, palletKey.."#i3dFilename");
            if i3dFilename then
                i3dFilename = Utils.convertFromNetworkFilename(i3dFilename);
			end;
            local xmlFilename = getXMLString(savegame.xmlFile, palletKey.."#xmlFilename");
            if xmlFilename then
                xmlFilename = Utils.convertFromNetworkFilename(xmlFilename);
			end;
			local fillLevel = Utils.getNoNil(getXMLFloat(savegame.xmlFile, palletKey.."#fillLevel"), 0);
			local percent = Utils.getNoNil(getXMLFloat(savegame.xmlFile, palletKey.."#percent"), 0);
			local fruitType;
			local fruitTypeName = getXMLString(savegame.xmlFile, palletKey.."#fruitType");
			if fruitTypeName then
				fruitType = FruitUtil.fruitTypes[fruitTypeName].index;
			end;
			local seedClass = getXMLInt(savegame.xmlFile, palletKey.."#seedClass");
			table.insert(self.palletLoader.loadPallets, {i3dFilename=i3dFilename, xmlFilename=xmlFilename, fillLevel=fillLevel,
														percent=percent, fruitType=fruitType, seedClass=seedClass});
			i = i + 1;
		end;
		self.palletLoader.sendUsedCount = #self.palletLoader.loadPallets;
	end;
end;

function PalletLoader:readStream(streamId, connection)
	local count = streamReadInt8(streamId);
	if count > 0 then
		self.palletLoader.loadPallets = {};
		for i=1, count do
			local i3dFilename = Utils.convertFromNetworkFilename(streamReadString(streamId));
			local percent = streamReadInt8(streamId)*0.01;
			table.insert(self.palletLoader.loadPallets, {i3dFilename=i3dFilename, percent=percent});
		end;
	end;
	self.palletLoader.state = streamReadInt8(streamId);
	self.palletLoader.grabMode = streamReadBool(streamId);
end;

function PalletLoader:writeStream(streamId, connection)
	local count = self.palletLoader.sendUsedCount;
	if not count then
		count = self.palletLoader.usedCount;
	end;
	streamWriteInt8(streamId, count);
	if count > 0 then
		for i=1, count do
			local i3dFilename = "";
			local percent = 1;
			local pos = self.palletLoader.positions[i];
			local loadData = self.palletLoader.loadPallets;
			if pos.pallet then
				i3dFilename = pos.pallet.i3dFilename;
				percent = pos.pallet.percent;
			elseif loadData then
				i3dFilename = loadData[i].i3dFilename;
				percent = loadData[i].percent;
			end;
			streamWriteString(streamId, Utils.convertToNetworkFilename(i3dFilename));
			streamWriteInt8(streamId, math.ceil(percent*100));
		end;
	end;
	streamWriteInt8(streamId, self.palletLoader.state);
	streamWriteBool(streamId, Utils.getNoNil(self.palletLoader.grabMode, false));
end;

function PalletLoader:getSaveAttributesAndNodes(nodeIdent)
    local attributes = "";
    local nodes = "";
	attributes = attributes..'palletLoaderState="'..string.format("%d", self.palletLoader.state)..'"';
	for i=1, #self.palletLoader.positions do
		local pos = self.palletLoader.positions[i];
		if pos.pallet then
			if i>1 then nodes = nodes.."\n"; end;
			nodes = nodes..nodeIdent..'<pallet i3dFilename="'..Utils.encodeToHTML(Utils.convertToNetworkFilename(pos.pallet.i3dFilename))..
			'" fillLevel="'..pos.pallet.fillLevel..'" percent="'..pos.pallet.percent..'"';
			if pos.pallet.fruitType and pos.pallet.seedClass then
				local fruitType = FruitUtil.fruitIndexToDesc[pos.pallet.fruitType].name;
				nodes = nodes..' fruitType="'..fruitType..'" seedClass="'..pos.pallet.seedClass..'"';
			end;
			if pos.pallet.xmlFilename then
				nodes = nodes..' xmlFilename="'..Utils.encodeToHTML(Utils.convertToNetworkFilename(pos.pallet.xmlFilename))..'"';
			end;
			nodes = nodes..' />';
		end;
	end;
	return attributes,nodes;
end;

function PalletLoader:update(dt)
	if self.palletLoader.loadPallets then
		for i=1, #self.palletLoader.loadPallets do
			local pallet = self.palletLoader.loadPallets[i];
			self:addPalletToPosition(i, pallet.i3dFilename, pallet.fillLevel, pallet.percent, pallet.xmlFilename, pallet.fruitType, pallet.seedClass);
			local position = self.palletLoader.positions[i];
			setTranslation(position.moveablePoint, 0,0,0);
			setRotation(position.pallet.palletNode, 0, math.rad(90*position.direction), 0);
		end;
		local startPos = self.palletLoader.startPosition;
		self.palletLoader.dropPosition = startPos[3] + 1.5*math.ceil(math.max(self.palletLoader.usedCount-2, 0)/2);
		self.palletLoader.lastDropPosition = self.palletLoader.dropPosition;
		setTranslation(self.palletLoader.palletGroup, startPos[1], startPos[2], self.palletLoader.dropPosition);
		if self.palletLoader.state ~= PalletLoader.STATE_CLOSED then
			self:onChangeState(PalletLoader.SENDSTATE_PREPARE);
		end;
		self.palletLoader.loadPallets = nil;
		self.palletLoader.sendUsedCount = nil;
	end;
	local rootAttacherVehicle = self:getRootAttacherVehicle();
	if self:getIsActive() and rootAttacherVehicle and rootAttacherVehicle.motor and rootAttacherVehicle:getIsMotorStarted() then
		if self.palletLoader.state == PalletLoader.STATE_CLOSED and self:getIsLoaderActive() then
			if InputBinding.hasEvent(self.palletLoader.foldInputButton) then
				if self.isServer then
					self:onChangeState(PalletLoader.SENDSTATE_PREPARE);
					g_server:broadcastEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_PREPARE, self));
				else
					g_client:getServerConnection():sendEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_PREPARE, self));
				end;
			end;
		end;
		if self.palletLoader.state == PalletLoader.STATE_READY then
			if self:getIsActiveForInput() and InputBinding.hasEvent(self.palletLoader.foldInputButton) then
				if self.isServer then
					self:onChangeState(PalletLoader.SENDSTATE_CLOSING);
					g_server:broadcastEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_CLOSING, self));
				else
					g_client:getServerConnection():sendEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_CLOSING, self));
				end;
			end;
			if InputBinding.hasEvent(self.palletLoader.loadInputButton) and self.palletLoader.usedCount < self.palletLoader.count then
				if self.isServer then
					self:onChangeState(PalletLoader.SENDSTATE_GRABMODE);
					g_server:broadcastEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_GRABMODE, self));
				else
					g_client:getServerConnection():sendEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_GRABMODE, self));
				end;
			end;
			if self.palletLoader.usedCount > 0 and InputBinding.hasEvent(self.palletLoader.unloadInputButton) then
				if self.isServer then
					self:onChangeState(PalletLoader.SENDSTATE_UNLOADING);
					g_server:broadcastEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_UNLOADING, self));
				else
					g_client:getServerConnection():sendEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_UNLOADING, self));
				end;
			end;
			if self.isServer and self.palletLoader.grabMode then
				local pos = self:getEmptyPosition();
				if pos then
					local position = self.palletLoader.positions[pos];
					local pallet = self:getPalletFromArea();
					if pallet then
						local fillLevel = pallet:getFillLevel();
						local percent = fillLevel/pallet:getCapacity();
						local fruitType, seedClass;
						if pallet.seedsExtension then
							fruitType = pallet.seedsExtension.fruitType;
							seedClass = pallet.seedsExtension.seedClass;
						end;
						self:addPalletToPosition(pos, pallet.i3dFilename, fillLevel, percent, pallet.xmlFilename, fruitType, seedClass);
						g_server:broadcastEvent(PalletLoaderLoadEvent:new(pos, pallet.i3dFilename, percent, self));
						self:onChangeState(PalletLoader.SENDSTATE_LOAD, pos);
						pallet:delete();
					end;
				else
					self.palletLoader.grabMode = false;
					g_server:broadcastEvent(PalletLoaderStateEvent:new(PalletLoader.SENDSTATE_GRABMODE, self));
				end;
			end;
		end;
		if self.palletLoader.state == PalletLoader.STATE_LOAD then
			local pos = self.palletLoader.loadingPosition;
			local x,y,z = worldToLocal(pos.node, getWorldTranslation(self.palletLoader.grabPoint));
			if self:getIsAnimationPlaying("liftOnLoad") then
				setTranslation(pos.moveablePoint, x,y,z);
			else
				self.palletLoader.currentMoving = math.max(self.palletLoader.currentMoving - dt*0.001, 0);
				setTranslation(pos.moveablePoint, x*self.palletLoader.currentMoving,y,z*self.palletLoader.currentMoving);
				setRotation(pos.pallet.palletNode, 0, math.rad(90)*(1-self.palletLoader.currentMoving)*pos.direction, 0);
				if self.palletLoader.currentMoving == 0 then
					self.palletLoader.loadingPosition = nil;
					self.palletLoader.state = PalletLoader.STATE_PREPARE;
					self:playAnimation("liftOnLoad", -1, nil, true);
					self.palletLoader.isFolding = true;
				end;
			end;
		end;
		if self.palletLoader.state == PalletLoader.STATE_BEFOREUNLOAD then
			local pos = self.palletLoader.loadingPosition;
			local x,y,z = worldToLocal(pos.node, getWorldTranslation(self.palletLoader.grabPoint));
			self.palletLoader.currentMoving = math.min(self.palletLoader.currentMoving + dt*0.001, 1);
			setTranslation(pos.moveablePoint, x*self.palletLoader.currentMoving,y,z*self.palletLoader.currentMoving);
			setRotation(pos.pallet.palletNode, 0, math.rad(90)*(1-self.palletLoader.currentMoving)*pos.direction, 0);
			if self.palletLoader.currentMoving == 1 then
				self.palletLoader.state = PalletLoader.STATE_UNLOAD;
				self:playAnimation("liftOnLoad", -1, nil, true);
				self.palletLoader.isFolding = true;
			end;
		end;
		if self.palletLoader.state == PalletLoader.STATE_UNLOAD then
			local pos = self.palletLoader.loadingPosition;
			local x,y,z = worldToLocal(pos.node, getWorldTranslation(self.palletLoader.grabPoint));
			if self:getIsAnimationPlaying("liftOnLoad") then
				setTranslation(pos.moveablePoint, x,y,z);
			else
				if self.isServer then
					local x,y,z = getWorldTranslation(self.palletLoader.unloadPoint);
					local _,ry,_ = getWorldRotation(self.palletLoader.unloadPoint);
					local i3dFilename = pos.pallet.i3dFilename;
					local palletType = "FillablePallet";
					if string.find(i3dFilename, "Sapling") then
						palletType = "SaplingPallet";
					end;
					local pallet = _G[palletType]:new(self.isServer, self.isClient);
					pallet:load(i3dFilename, x,y,z, 0,ry,0, pos.pallet.xmlFilename);
					pallet:register();
					pallet:setFillLevel(pos.pallet.fillLevel, true);
					if pos.pallet.fruitType and pos.pallet.seedClass then
						pallet.seedsExtension = {fruitType=pos.pallet.fruitType, seedClass=pos.pallet.seedClass};
						if FillablePallet.setFruitPicture then
							FillablePallet.setFruitPicture(pallet);
						end;
					end;
					pallet.sendI3dFilename = true;
					self:setFillLevel(self.palletLoader.usedCount - 1, FillUtil.FILLTYPE_PALLETS);
				end;
				self:removePalletFromPosition(pos);
				self.palletLoader.loadingPosition = nil;
				if self.palletLoader.usedCount > 1 then
					if pos.direction == 1 then
						self.palletLoader.dropMode = true;
						self.palletLoader.dropDirection = -1;
						self.palletLoader.currentPosition = 1;
						self.palletLoader.lastDropPosition = self.palletLoader.dropPosition - 1.5;
						self.palletLoader.state = PalletLoader.STATE_PREPARE;
					else
						self.palletLoader.state = PalletLoader.STATE_READY;
					end;
				else
					self.palletLoader.state = PalletLoader.STATE_READY;
				end;
			end;
		end;
		if self.palletLoader.dropMode then
			if self.palletLoader.dropDirection == 1 then
				self.palletLoader.currentPosition = math.min(self.palletLoader.currentPosition + dt*0.001, 1);
			else
				self.palletLoader.currentPosition = math.max(self.palletLoader.currentPosition - dt*0.001, 0);
			end;
			self.palletLoader.dropPosition = self.palletLoader.lastDropPosition + 1.5*self.palletLoader.currentPosition;
			local startPos = self.palletLoader.startPosition;
			setTranslation(self.palletLoader.palletGroup, startPos[1], startPos[2], self.palletLoader.dropPosition);
			if (self.palletLoader.dropDirection == 1 and self.palletLoader.currentPosition == 1) or
			(self.palletLoader.dropDirection == -1 and self.palletLoader.currentPosition == 0) then
				self.palletLoader.dropMode = false;
				self.palletLoader.lastDropPosition = self.palletLoader.dropPosition;
			end;
		end;
		if self.palletLoader.isFolding and not self:getIsAnimationPlaying("liftOnLoad") and not self:getIsAnimationPlaying("folding") then
			self.palletLoader.isFolding = false;
		end;
		if self.palletLoader.state == PalletLoader.STATE_PREPARE and not self:getIsAnimationPlaying("liftOnLoad") and not self:getIsAnimationPlaying("folding") then
			if self.palletLoader.loadingPosition then
				self.palletLoader.state = PalletLoader.STATE_BEFOREUNLOAD;
			else
				self.palletLoader.state = PalletLoader.STATE_READY;
			end;
		end;
		if self.palletLoader.state == PalletLoader.STATE_CLOSING and not self:getIsAnimationPlaying("folding") then
			self.palletLoader.state = PalletLoader.STATE_CLOSED;
		end;
	end;
end;

function PalletLoader:updateTick(dt)
end;

function PalletLoader:draw()
	if self:getIsLoaderActive() then
		if self.palletLoader.state == PalletLoader.STATE_CLOSED then
			g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_unfoldOBJECT"), self.typeDesc), self.palletLoader.foldInputButton, nil, GS_PRIO_HIGH);
		end;
		if self.palletLoader.state == PalletLoader.STATE_READY then
			g_currentMission:addHelpButtonText(string.format(g_i18n:getText("action_foldOBJECT"), self.typeDesc), self.palletLoader.foldInputButton, nil, GS_PRIO_HIGH);
			if self.palletLoader.usedCount < self.palletLoader.count then
				g_currentMission:addHelpButtonText(self.palletLoader.loadText[self.palletLoader.grabMode], self.palletLoader.loadInputButton, nil, GS_PRIO_HIGH);
			end;
			if self.palletLoader.usedCount > 0 then
				g_currentMission:addHelpButtonText(g_i18n:getText("action_unloadPallet"), self.palletLoader.unloadInputButton, nil, GS_PRIO_HIGH);
			end;
		end;
	end;
end;

function PalletLoader:getIsLoaderActive()
	if self:getIsActiveForInput(true) then
		if self.palletLoader.lockJointIndex then
			if self.attachedImplements and #self.attachedImplements > 0 then
				local enable = true;
				for _,implement in pairs(self.attachedImplements) do
					if implement.jointDescIndex == self.palletLoader.lockJointIndex then
						enable = false;
						break;
					end;
				end;
				return enable;
			else
				return true;
			end;
		else
			return true;
		end;
	end;
	return false;
end;

function PalletLoader:getPalletFromArea()
	if g_currentMission.itemsToSave then
		for index,item in pairs(g_currentMission.itemsToSave) do
			local pallet = item.item;
			if pallet:isa(FillablePallet) then
				local x, _, z = worldToLocal(self.palletLoader.startPoint, getWorldTranslation(pallet.nodeId));
				if x > 0 and z < 0 and x < self.palletLoader.dimensions[1] and z > self.palletLoader.dimensions[2] then
					return pallet;
				end;
			end;
		end;
	end;
end;

function PalletLoader:getEmptyPosition()
	for i=1, #self.palletLoader.positions do
		local pos = self.palletLoader.positions[i];
		if pos.pallet == nil then
			return i;
		end;
	end;
end;

function PalletLoader:getLastUsedPosition()
	if self.palletLoader.usedCount > 0 then
		return self.palletLoader.positions[self.palletLoader.usedCount];
	end;
end;

function PalletLoader:getUnloadingPlaceIsEmpty()
	if g_currentMission.itemsToSave then
		for _,item in pairs(g_currentMission.itemsToSave) do
			local obj = item.item;
			if obj:isa(FillablePallet) or obj:isa(Bale) then
				local x,_,z = worldToLocal(self.palletLoader.unloadPoint, getWorldTranslation(obj.nodeId));
				local distance = Utils.vector2Length(x, z);
				if distance < 1 then
					g_currentMission:showBlinkingWarning(g_i18n:getText("otherObject"), 2000);
					return false;
				end;
			end;
		end;
	end;
	return true;
end;

function PalletLoader:addPalletToPosition(pos, i3dFilename, fillLevel, percent, xmlFilename, fruitType, seedClass)
	local position = self.palletLoader.positions[pos];
	local palletRoot = Utils.loadSharedI3DFile(i3dFilename);
	local palletNode = getChildAt(palletRoot, 0);
	removeFromPhysics(palletNode);
	setRigidBodyType(palletNode, "NoRigidBody");
	setScale(palletNode, 1, 1, 0.98);
	link(position.moveablePoint, palletNode);
	delete(palletRoot);
	if percent then
		local visibilityNodes = Utils.indexToObject(palletNode, getUserAttribute(palletNode, "visibilityNodesIndex"));
		local visibilityNodesAdditive = Utils.getNoNil(getUserAttribute(palletNode, "visibilityNodesAdditive"), false);
		local saplingFilename = getUserAttribute(palletNode, "saplingFilename");
		if visibilityNodes ~= nil then
			local numVisibilityNodes = getNumOfChildren(visibilityNodes);
			local numVisible = math.ceil(numVisibilityNodes*percent);
			if saplingFilename then
				local i3dNode = Utils.loadSharedI3DFile(saplingFilename);
				if i3dNode > 0 then
					for i=1, numVisible do
						local node = clone(i3dNode, false, false, false);
						setScale(node, 1, 0.95, 1);
						link(getChildAt(visibilityNodes, i-1), node);
						if getUserAttribute(palletNode, "useRandomRot") then
							setRotation(node, 0, math.random(0, 2*math.pi), 0);
						end;
					end;
				end;
				delete(i3dNode);
			else
				if visibilityNodesAdditive then
					for i=1,numVisibilityNodes do
						setVisibility(getChildAt(visibilityNodes, i-1), i <= numVisible);
					end;
				else
					for i=1,numVisibilityNodes do
						setVisibility(getChildAt(visibilityNodes, i-1), i == numVisible);
					end;
				end;
			end;
		end;
	end;
	local effectsNode = getChild(palletNode, "effects");
	if effectsNode > 0 then
		delete(effectsNode);
	end;
	position.pallet = {	palletNode=palletNode, i3dFilename=i3dFilename, fillLevel=fillLevel,
						percent=percent, xmlFilename=xmlFilename, fruitType=fruitType, seedClass=seedClass};
	self.palletLoader.usedCount = self.palletLoader.usedCount + 1;
	if self.isServer then
		self:setFillLevel(self.palletLoader.usedCount, FillUtil.FILLTYPE_PALLETS);
	end;
end;

function PalletLoader:removePalletFromPosition(position)
	delete(position.pallet.palletNode);
	position.pallet = nil;
	self.palletLoader.usedCount = self.palletLoader.usedCount - 1;
end;

function PalletLoader:onChangeState(sendState, pos)
	if sendState == PalletLoader.SENDSTATE_PREPARE then
		self:playAnimation("folding", 1, nil, true);
		self.palletLoader.isFolding = true;
		self.palletLoader.state = PalletLoader.STATE_PREPARE;
	elseif sendState == PalletLoader.SENDSTATE_CLOSING then
		self.palletLoader.grabMode = false;
		self:playAnimation("folding", -1, nil, true);
		self.palletLoader.isFolding = true;
		self.palletLoader.state = PalletLoader.STATE_CLOSING;
	elseif sendState == PalletLoader.SENDSTATE_LOAD then
		local position = self.palletLoader.positions[pos];
		if position then
			self.palletLoader.loadingPosition = position;
			local x,y,z = worldToLocal(position.node, getWorldTranslation(self.palletLoader.grabPoint));
			setTranslation(position.moveablePoint, x,y,z);
			self:playAnimation("liftOnLoad", 1, nil, true);
			self.palletLoader.isFolding = true;
			self.palletLoader.currentMoving = 1;
			if self.palletLoader.usedCount < self.palletLoader.count and self.palletLoader.usedCount > 1 and position.direction == 1 then
				self.palletLoader.dropMode = true;
				self.palletLoader.dropDirection = 1;
				self.palletLoader.currentPosition = 0;
			end;
			self.palletLoader.state = PalletLoader.STATE_LOAD;
		end;
	elseif sendState == PalletLoader.SENDSTATE_UNLOADING then
		if self:getUnloadingPlaceIsEmpty() then
			self.palletLoader.grabMode = false;
			self.palletLoader.loadingPosition = self:getLastUsedPosition();
			self:playAnimation("liftOnLoad", 1, nil, true);
			self.palletLoader.isFolding = true;
			self.palletLoader.currentMoving = 0;
			self.palletLoader.state = PalletLoader.STATE_PREPARE;
		end;
	elseif sendState == PalletLoader.SENDSTATE_GRABMODE then
		self.palletLoader.grabMode = not self.palletLoader.grabMode;
	end;
end;

function PalletLoader:getDoConsumePtoPower()
	return self.palletLoader.isFolding;
end;

function PalletLoader:getPtoRpm()
	if self.palletLoader.isFolding then
        return self.powerConsumer.ptoRpm;
    end;
    return 0;
end;

-----------------------------------------------------------------------

PalletLoaderStateEvent = {};
PalletLoaderStateEvent_mt = Class(PalletLoaderStateEvent, Event);
InitEventClass(PalletLoaderStateEvent, "PalletLoaderStateEvent");

function PalletLoaderStateEvent:emptyNew()
	local self = Event:new(PalletLoaderStateEvent_mt);
    return self;
end;

function PalletLoaderStateEvent:new(num, loader)
	local self = PalletLoaderStateEvent:emptyNew();
	self.num = num;
	self.loader = loader;
	return self;
end;

function PalletLoaderStateEvent:readStream(streamId, connection)
	local num = streamReadInt8(streamId);
	local id = streamReadInt32(streamId);
    local loader = networkGetObject(id);
	if loader then
		loader:onChangeState(num);
		if g_server then
			g_server:broadcastEvent(PalletLoaderStateEvent:new(num, loader));
		end;
	end;
end;

function PalletLoaderStateEvent:writeStream(streamId, connection)
	streamWriteInt8(streamId, self.num);
	streamWriteInt32(streamId, networkGetObjectId(self.loader));
end;

-----------------------------------------------------------------------

PalletLoaderLoadEvent = {};
PalletLoaderLoadEvent_mt = Class(PalletLoaderLoadEvent, Event);
InitEventClass(PalletLoaderLoadEvent, "PalletLoaderLoadEvent");

function PalletLoaderLoadEvent:emptyNew()
	local self = Event:new(PalletLoaderLoadEvent_mt);
    return self;
end;

function PalletLoaderLoadEvent:new(position, i3dFilename, percent, loader)
	local self = PalletLoaderLoadEvent:emptyNew();
	self.position = position;
	self.i3dFilename = i3dFilename;
	self.percent = percent;
	self.loader = loader;
	return self;
end;

function PalletLoaderLoadEvent:readStream(streamId, connection)
	local position = streamReadInt8(streamId);
	local i3dFilename = Utils.convertFromNetworkFilename(streamReadString(streamId));
	local percent = streamReadInt16(streamId)*0.01;
	local id = streamReadInt32(streamId);
    local loader = networkGetObject(id);
	if loader then
		loader:addPalletToPosition(position, i3dFilename, nil, percent);
		loader:onChangeState(PalletLoader.SENDSTATE_LOAD, position);
	end;
end;

function PalletLoaderLoadEvent:writeStream(streamId, connection)
	streamWriteInt8(streamId, self.position);
	streamWriteString(streamId, Utils.convertToNetworkFilename(self.i3dFilename));
	streamWriteInt16(streamId, math.ceil(self.percent)*100);
	streamWriteInt32(streamId, networkGetObjectId(self.loader));
end;
