--
-- emptyMe
-- Specialization for empty an fillable object
--
-- by Blacky_BPG
-- 
-- Version: 1.4.4.0      |    07.06.2017    initial Version for FS17
-- 

emptyMe = {}

function emptyMe.prerequisitesPresent(specializations) return SpecializationUtil.hasSpecialization(Fillable, specializations) end

function emptyMe:load(xmlFile)
	self.showTipOption = 0
	self.tippingToEmpty = false
	self.emptyTipSound = false
	self.emptyingMyLoad = SpecializationUtil.callSpecializationsFunction("emptyingMyLoad")
end
function emptyMe:update(dt)
	if self:getIsActiveForInput() then
		if self.showTipOption > 0 then
			if InputBinding.hasEvent(InputBinding.EMPTY_ME) then
				self.tippingToEmpty = not self.tippingToEmpty
				emptyMeEvent.sendEvent(self, self.tippingToEmpty)
			end
		end
	end
end
function emptyMe:updateTick(dt)
	if self:getIsActiveForInput() then
		self.showTipOption = 0
		for i=1,table.getn(self.fillUnits) do
			local fillUnit = self.fillUnits[i]
			if fillUnit.fillLevel > 0 then
				self.showTipOption = i
				break
			end
		end
	end
	if self.isServer then
		if self.tippingToEmpty then
			self.showTipOption = 0
			for i=1,table.getn(self.fillUnits) do
				local fillUnit = self.fillUnits[i]
				if fillUnit.fillLevel > 0 then
					self.showTipOption = i
					break
				end
			end
			if self.showTipOption <= 0 then
				self.tippingToEmpty = false
				emptyMeEvent.sendEvent(self, false)
			else
				self:emptyingMyLoad(dt)
			end
		end
	end
	if self.isClient then
		if self.tippingToEmpty then
			if self.sampleFillSound ~= nil then
				if self:getIsActiveForSound() then
					if not self.emptyTipSound then
						SoundUtil.playSample(self.sampleFillSound, 0, 0, nil)
					end
					self.emptyTipSound = true
				else
					if self.emptyTipSound then
						SoundUtil.stopSample(self.sampleFillSound)
						self.emptyTipSound = false
					end
				end
			end
		else
			if self.sampleFillSound ~= nil and self.emptyTipSound then
				SoundUtil.stopSample(self.sampleFillSound)
				self.emptyTipSound = false
			end
		end
	end
end
function emptyMe:emptyingMyLoad(dt)
	if self.isServer then
		if self.tippingToEmpty and self.showTipOption > 0 then
			local fillLevel = self:getUnitFillLevel(self.showTipOption)
			fillLevel = fillLevel - (Utils.getNoNil(self.fillLitersPerSecond,self:getUnitCapacity(self.showTipOption)/20)*dt*0.001)
			self:setUnitFillLevel(self.showTipOption, fillLevel, self:getUnitFillType(self.showTipOption), true)
		end
	end
end

function emptyMe:mouseEvent(posX, posY, isDown, isUp, button) end
function emptyMe:keyEvent(unicode, sym, modifier, isDown) end
function emptyMe:readStream(streamId, connection) end 
function emptyMe:readUpdateStream(streamId, timestamp, connection) end 
function emptyMe:writeStream(streamId, connection) end 
function emptyMe:writeUpdateStream(streamId, connection, dirtyMask) end 
function emptyMe:draw()
	if self:getIsActiveForInput() then
		if self.showTipOption > 0 then
			if self.tippingToEmpty then
				g_currentMission:addHelpButtonText(g_i18n:getText("input_EMPTY_ME_OFF"), InputBinding.EMPTY_ME, nil, GS_PRIO_HIGH)
			else
				g_currentMission:addHelpButtonText(g_i18n:getText("input_EMPTY_ME_ON"), InputBinding.EMPTY_ME, nil, GS_PRIO_HIGH)
			end
		end
	end
end
function emptyMe:delete() end

-----------------------------------------E V E N T-------------------------------------------------

emptyMeEvent = {}
emptyMeEvent_mt = Class(emptyMeEvent, Event)

InitEventClass(emptyMeEvent, "emptyMeEvent")

function emptyMeEvent:emptyNew()
    local self = Event:new(emptyMeEvent_mt)
    self.className="emptyMeEvent"
    return self
end

function emptyMeEvent:new(vehicle, state)
    local self = emptyMeEvent:emptyNew()
    self.vehicle = vehicle
	self.state = state
	return self
end

function emptyMeEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId)
	self.state = streamReadBool(streamId)
    self.vehicle = networkGetObject(id)
    self:run(connection)
end

function emptyMeEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.vehicle))
	streamWriteBool(streamId, self.state)
end

function emptyMeEvent:run(connection)
	self.vehicle.tippingToEmpty = self.state
	if not connection:getIsServer() then
		g_server:broadcastEvent(emptyMeEvent:new(self.vehicle, self.state), nil, connection, self.object)
	end
end

function emptyMeEvent.sendEvent(vehicle, state, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(emptyMeEvent:new(vehicle, state), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(emptyMeEvent:new(vehicle, state))
		end
	end
end
